from typing import Any


class EntityTypeException(Exception):
    """The value does not correspond to the abc class SocialMediaEntity"""

    def __init__(self, message: Any) -> None:
        self.message = (
            f"{message} does not correspond to the abc class SocialMediaEntity"
        )

    def __str__(self) -> str:
        return self.message


class PresenceObjectException(Exception):
    """There are no entities of the passed type"""

    def __init__(self, message: Any) -> None:
        self.message = f"There is no essential {message} type"

    def __str__(self) -> str:
        return self.message


class BanningAbcClass(Exception):
    """It is forbidden to transmit abc SocialMediaEntity"""

    def __str__(self) -> str:
        return "It is forbidden to transmit abc SocialMediaEntity"



import uuid

from foxy_entities import SocialMediaEntity


class TestSocialMediaEntity(SocialMediaEntity):
    test_str: str


class TestSocialMediaEntity2(SocialMediaEntity):
    test_str2: str


def test_id_social_media_entity():
    test_str = TestSocialMediaEntity(test_str="test_str")
    assert type(test_str.id) is uuid.UUID



from typing import Self, Type, TypeVar

from foxy_entities.abc import SocialMediaEntity
from foxy_entities.exceptions import (
    EntityTypeException,
    PresenceObjectException,
    BanningAbcClass,
)

SocialMediaEntityType = TypeVar("SocialMediaEntityType", bound="SocialMediaEntity")


class EntitiesController:
    """
    A virtual storage controller class that provides methods
    for adding and retrieving entities with simple Round Robin balancing
    """

    def __init__(self) -> None:
        self.__entity_virtual_storage: dict[str, list[SocialMediaEntity]] = {}

    def get_virtual_storage(self) -> dict[str, list[SocialMediaEntity]]:
        return self.__entity_virtual_storage.copy()

    @staticmethod
    def update_sequence_entity(
        sequence_entity: list[SocialMediaEntity], social_media_entity: SocialMediaEntity
    ) -> list[SocialMediaEntity]:
        sequence_entity = [social_media_entity] + sequence_entity
        return sequence_entity

    def add_entity(self, social_media_entity: SocialMediaEntity) -> Self:
        if not isinstance(social_media_entity, SocialMediaEntity):
            raise EntityTypeException(social_media_entity)
        if type(social_media_entity) is SocialMediaEntity:
            raise BanningAbcClass()
        sequence_entity = self.__entity_virtual_storage.get(
            social_media_entity.__class__.__name__
        )
        if sequence_entity is not None:
            self.__entity_virtual_storage[social_media_entity.__class__.__name__] = (
                self.update_sequence_entity(sequence_entity, social_media_entity)
            )
        else:
            self.__entity_virtual_storage[social_media_entity.__class__.__name__] = [
                social_media_entity
            ]
        return self

    def get_entity(
        self, social_media_type: Type[SocialMediaEntityType]
    ) -> SocialMediaEntityType:
        if not issubclass(social_media_type, SocialMediaEntity):
            raise EntityTypeException(social_media_type)
        if social_media_type is SocialMediaEntity:
            raise BanningAbcClass()

        sequence_entity = self.__entity_virtual_storage.get(social_media_type.__name__)

        if not sequence_entity:
            raise PresenceObjectException(social_media_type)

        social_media_entity = sequence_entity.pop()

        if sequence_entity:
            self.__entity_virtual_storage[social_media_type.__name__] = sequence_entity
        else:
            del self.__entity_virtual_storage[social_media_type.__name__]

        return social_media_entity



import pytest

from foxy_entities import EntitiesController, SocialMediaEntity
from foxy_entities.exceptions import (
    EntityTypeException,
    BanningAbcClass,
    PresenceObjectException,
)
from tests.test_abc import TestSocialMediaEntity


@pytest.mark.parametrize(
    "input_value",
    [
        1,  # int
        0.2,  # float
        [1, 4, 5],  # list
        "hello",  # str
        True,  # bool
        None,  # NoneType
        (1, 2, 3),  # tuple
        {"key": "value"},  # dict
        {1, 2, 3},  # set
        3 + 4j,  # complex
        b"bytes",  # bytes
        bytearray(b"test"),  # bytearray
    ],
)
def test_exception_entity_type_add_entity(input_value):
    """
    Test case for checking the prohibition of transmission of
    a type that does not correspond to SocialMediaEntity
    """
    with pytest.raises(EntityTypeException):
        EntitiesController().add_entity(input_value)


def test_add_entity_incorrect_type_abc_class():
    """
    Test case prohibiting transfer of abc class
    """
    with pytest.raises(BanningAbcClass):
        EntitiesController().add_entity(SocialMediaEntity())


def test_add_entity_position():
    """
    Test case of how getting an entity in a FIFO queue works
    """
    entity_controller = EntitiesController()
    test_entity_1 = TestSocialMediaEntity(test_str="test_entity_1")
    test_entity_2 = TestSocialMediaEntity(test_str="test_entity_2")
    entity_controller.add_entity(test_entity_1).add_entity(test_entity_2)
    received_entity = entity_controller.get_entity(TestSocialMediaEntity)
    assert test_entity_1 == received_entity
    received_entity = entity_controller.get_entity(TestSocialMediaEntity)
    assert test_entity_2 == received_entity


def test_unique_sequences_objects():
    """
    Test case virtual storages should not overlap between different objects
    """
    entity_controller_1 = EntitiesController()
    entity_controller_2 = EntitiesController()
    test_entity_1 = TestSocialMediaEntity(test_str="test_entity_1")
    test_entity_2 = TestSocialMediaEntity(test_str="test_entity_2")
    entity_controller_1.add_entity(test_entity_1)
    entity_controller_2.add_entity(test_entity_2)
    assert (
        entity_controller_1.get_virtual_storage()
        != entity_controller_2.get_virtual_storage()
    )


@pytest.mark.parametrize(
    "input_value",
    [
        1,  # int
        0.2,  # float
        [1, 4, 5],  # list
        "hello",  # str
        True,  # bool
        None,  # NoneType
        (1, 2, 3),  # tuple
        {"key": "value"},  # dict
        {1, 2, 3},  # set
        3 + 4j,  # complex
        b"bytes",  # bytes
        bytearray(b"test"),  # bytearray
    ],
)
def test_get_entity_incorrect_type_base(input_value):
    """
    Test case virtual storages should not overlap between different objects
    """
    with pytest.raises(TypeError):
        EntitiesController().get_entity(input_value)


def test_get_entity_incorrect_type_abc_class():
    """
    Test case prohibiting transfer of abc class
    """
    with pytest.raises(BanningAbcClass):
        EntitiesController().get_entity(SocialMediaEntity)


def test_get_entity_out_stock():
    """
    Test case exceptions when the passed type is not in the virtual storage
    """
    with pytest.raises(PresenceObjectException):
        EntitiesController().get_entity(TestSocialMediaEntity)


def test_purge_virtual_storage():
    """
    Test case virtual storage must clear empty entity lists
    """
    entity_controller = EntitiesController()
    entity_controller.add_entity(TestSocialMediaEntity(test_str="test_entity_1"))
    entity_controller.get_entity(TestSocialMediaEntity)
    assert entity_controller.get_virtual_storage() == {}


def test_independence_virtual_storage():
    """
    Test case virtual storage must clear empty entity lists
    """
    entity_controller = EntitiesController()
    entity_controller.add_entity(TestSocialMediaEntity(test_str="test_entity_1"))
    entity_copy = entity_controller.get_virtual_storage()
    entity_copy["TestSocialMediaEntity2"] = [
        TestSocialMediaEntity(test_str="test_entity_2")
    ]
    assert entity_copy != entity_controller.get_virtual_storage()
